<?php
/**
 * Builder Test
 *
 * @package Holograph
 */

namespace Holograph\Test;

use \BaseTestCase;
use Holograph\Builder;
use Holograph\Logger\Memory;

/**
 * BuilderTest
 *
 * @uses BaseTestCase
 * @package Holograph
 * @author Jansen Price <jansen.price@gmail.com>
 * @version $Id$
 */
class BuilderTest extends BaseTestCase
{
    /**
     * Logging object
     *
     * @var mixed
     */
    public $logger;

    /**
     * Set up before test
     *
     * @return void
     */
    public function setUp()
    {
        $this->logger = new Memory();

        $this->_object = new Builder(array(), $this->logger);
    }

    /**
     * Tear down after test
     *
     * @return void
     */
    public function tearDown()
    {
        // clean up
        passthru("rm -rf test-source");
        passthru("rm -rf test-destination");
        passthru("rm -rf test-build");
    }

    /**
     * Test constructor
     *
     * @return void
     * @expectedException PHPUnit_Framework_Error_Warning
     */
    public function testConstruct()
    {
        $builder = new Builder();
    }

    /**
     * Test constructor with empty array as config
     *
     * @return void
     * @expectedException PHPUnit_Framework_Error_Warning
     */
    public function testConstructWithConfigEmpty()
    {
        $builder = new Builder(array());
    }

    /**
     * Test constructor with invalid object
     *
     * @return void
     * @expectedException PHPUnit_Framework_Error
     */
    public function testConstructWithIncorrectClient()
    {
        $c = new \StdClass();

        $builder = new Builder(array(), $c);
    }

    /**
     * testConstructWithMockLogger
     *
     * @return void
     */
    public function testConstructWithMockLogger()
    {
        $argv = $this->getMock("Qi_Console_ArgV", array(), array(array('foo')));

        $terminal = $this->getMock("Qi_Console_Terminal");

        $logger = $this->getMock(
            "Holograph\\Logger\\Terminal", array(), array($terminal)
        );

        $logger->expects($this->once())
            ->method("info")
            ->will($this->returnValue(null));

        $builder = new Builder(array("a: b"), $logger);
    }

    /**
     * testGetConfig
     *
     * @return void
     */
    public function testGetConfig()
    {
        $config = $this->_object->getConfig();

        $expected = array(
            'title'                => "Style Guide",
            'source'               => "./components",
            'destination'          => "./docs",
            'documentation_assets' => "./templates",
            'compat_mode'          => false,
            'dependencies'         => array("./build"),
            'preprocessor'         => "minify",
            'build'                => "./build/css",
            'main_stylesheet'      => "build/css/screen.css",
        );

        $this->assertEquals($expected, $config);
    }

    /**
     * testGetConfigWhenAddedTo
     *
     * @return void
     */
    public function testGetConfigWhenAddedTo()
    {
        $config = array('source' => 'FFFFFFFFF');

        $builder = new Builder($config, $this->logger);

        $expected = 'FFFFFFFFF';

        $resultConfig = $builder->getConfig();

        $actual = $resultConfig['source'];

        $this->assertEquals($expected, $actual);
    }

    /**
     * testGetConfigAnnotated
     *
     * @return void
     */
    public function testGetConfigAnnotated()
    {
        $doc = $this->_object->getConfigAnnotated();

        $this->assertContains("Holograph configuration", $doc);
        $this->assertContains("Directory to build the final", $doc);
    }

    /**
     * testExecute
     *
     * @return void
     */
    public function testExecute()
    {
        $result = $this->_object->execute();

        $this->assertEquals(1, $result);
    }

    /**
     * testExecuteSimple
     *
     * @return void
     */
    public function testExecuteSimple()
    {
        mkdir("test-source");
        file_put_contents("test-source/test.css", ".main { color: red}");
        file_put_contents("test-source/test.md", "# Hi there");

        $config = array(
            'source' => 'test-source',
            'destination' => 'test-destination',
            'build' => 'test-build/css',
            'main_stylesheet' => 'test-build/css/screen.css',
        );

        $builder = new Builder($config, $this->logger);

        $result = $builder->execute();

        $this->assertTrue(file_exists("test-destination/static/css/doc.css"));
        $this->assertTrue(file_exists("test-build/css/screen.css"));
        $this->assertTrue(file_exists("test-destination/test.html"));
        $this->assertEquals(0, $result);
    }

    /**
     * testParseSourceFileWithoutYml
     *
     * @return void
     */
    public function testParseSourceFileWithoutYml()
    {
        mkdir("test-source");
        file_put_contents("test-source/test.css", "/*doc\nfoobar\n*/");

        $config = array(
            'source' => 'test-source',
            'destination' => 'test-destination',
            'build' => 'test-build/css',
            'main_stylesheet' => 'test-build/css/screen.css',
        );

        $builder = new Builder($config, $this->logger);

        $added = $builder->parseSourceStylesheetFile('test-source/test.css');

        $this->assertEquals(0, $added);
    }

    /**
     * testParseSourceFileWithYml
     *
     * @return void
     */
    public function testParseSourceFileWithYml()
    {
        mkdir("test-source");
        file_put_contents(
            "test-source/test.css", "/*doc\n---\nfoobar\n---\n*/"
        );

        $config = array(
            'source' => 'test-source',
            'destination' => 'test-destination',
            'build' => 'test-build/css',
            'main_stylesheet' => 'test-build/css/screen.css',
        );

        $builder = new Builder($config, $this->logger);

        $added = $builder->parseSourceStylesheetFile('test-source/test.css');

        $this->assertEquals(1, $added);
    }

    /**
     * testRunPreprocessorWithNone
     *
     * @return void
     */
    public function testRunPreprocessorWithNone()
    {
        $config = array(
            'preprocessor' => 'none',
        );

        $builder = new Builder($config, $this->logger);

        $result = $builder->runPreprocessor(array());
        $this->assertNull($result);
    }

    /**
     * testCreateDocumentBlockNoMatch
     *
     * @return void
     */
    public function testCreateDocumentBlockNoMatch()
    {
        $result = $this->_object->createDocumentBlock('abc', 'def');
        $this->assertFalse($result);
    }

    /**
     * testCreateDocumentBlockWithYml
     *
     * @return void
     */
    public function testCreateDocumentBlockWithYml()
    {
        $block = "---\nname: foobar\n---";

        $result = $this->_object->createDocumentBlock($block, 'def');

        $this->assertEquals("Holograph\\DocumentBlock", get_class($result));
    }

    /**
     * testCreateDocumentBlockWithInvalidYml
     *
     * @return void
     */
    public function testCreateDocumentBlockWithInvalidYml()
    {
        $block = "---\nhi\n---";

        $result = $this->_object->createDocumentBlock($block, 'def');

        $this->assertEquals("Holograph\\DocumentBlock", get_class($result));
    }
}
